#ifndef HYPER_GENOM_MODEL_COMPUTE_EXPRESSION_HH_
#define HYPER_GENOM_MODEL_COMPUTE_EXPRESSION_HH_

#include <model/update_impl.hh>

#include <genom_model/reference.hh>

namespace hyper {
	namespace genom_model {

		typedef boost::function<void (const boost::system::error_code& e)> cb_type;

		template <typename T>
		struct storage { typedef T type; };

		template <> struct storage<void> { typedef boost::mpl::void_ type; };

		template <typename Fun>
		class compute_expression {
			enum state { IDLE, RUNNING, ABORTING };

			public:
				typedef typename Fun::updater_type updater_type;
				typedef typename Fun::output_type output_type;
				typedef typename reference<output_type>::type output_ref_type;
				typedef typename Fun::genom_input_type genom_input_type;
				typedef typename Fun::genom_output_type genom_output_type;
	
			private:
				updater_type updater;
				boost::optional<ssize_t> rqstId;
				state state_;
				typename storage<genom_input_type>::type genom_input;
				typename storage<genom_output_type>::type genom_output;

				template <typename Handler>
				void handle_exec(const boost::system::error_code& e,
								  boost::tuple<Handler> handler)
				{
					rqstId = boost::none;
					state_ = IDLE;
					return boost::get<0>(handler)(e);
				}

				template <typename Handler, typename Ability, typename Local>
				void handle_update(const boost::system::error_code& e,
								   Ability& ability,
								   const Local& local,
								   output_ref_type output,
								   boost::tuple<Handler> handler)
				{
					if (e) 
						return boost::get<0>(handler)(e);
	
					void (compute_expression::*f) (const boost::system::error_code& e,
												   boost::tuple<Handler> handler) =
						&compute_expression::template handle_exec<Handler>;

					state_ = RUNNING;
					rqstId = Fun()(updater, ability, local, genom_input, genom_output, output, 
								   boost::bind(f, this, boost::asio::placeholders::error, handler));
				}
	
			public:
				compute_expression(updater_type updater) : updater(updater) {}
	
				template <typename Handler, typename Ability, typename Local>
				void async_eval(Handler handler, Ability& a, const Local& local,
								output_ref_type output)
				{
					rqstId = boost::none;
					state_ = IDLE;

					void (compute_expression::*f) (
							const boost::system::error_code&e,
							Ability& ,
							const Local& ,
							output_ref_type,
							boost::tuple<Handler>) =
						&compute_expression::template handle_update<Handler, Ability, Local>;
	
					updater.async_update(
							boost::bind(f, this, boost::asio::placeholders::error,
												 boost::ref(a), 
												 boost::cref(local),
												 boost::ref(output),
												 boost::make_tuple(handler)));
				}

				template <typename Ability, typename Local>
				bool abort(Ability& a, const Local& local)
				{
					switch(state_) {
						case RUNNING:
							Fun::abort(a, local, *rqstId);
							rqstId = boost::none;
							state_ = ABORTING;
							return true;
						case ABORTING:
							return true;
						default:	
							return false;
					}
				}
		};
	}
}

#endif /* HYPER_GENOM_MODEL_COMPUTE_EXPRESSION_HH_ */
